import codecs
import json
import base64
import mutagen
import mutagen.mp4
import mutagen.id3
from urllib.request import urlopen
import os
import platform
import re
import shutil
import subprocess
import sys
import tempfile

stemDescription  = 'stem-meta'
stemOutExtension = ".mp4"

_windows = platform.system() == "Windows"

_supported_files_no_conversion = [".m4a", ".mp4", ".m4p"]
_supported_files_conversion = [".wav", ".wave", ".aif", ".aiff", ".flac"]
_supported_files = _supported_files_no_conversion + _supported_files_conversion


def _removeFile(path):
    if os.path.lexists(path):
        if os.path.isfile(path):
            os.remove(path)
        else:
            raise RuntimeError("Cannot remove " + path + ": not a file")


def _getProgramPath():
    folderPath = os.path.dirname(os.path.realpath(__file__))
    if os.path.isfile(folderPath):
        # When packaging the script with py2exe, os.path.realpath returns the path to the object file within
        # the .exe, so we have to apply os.path.dirname() one more time
        folderPath = os.path.dirname(folderPath)
    return folderPath


def _make_temp_unique_path(file_name):
    # Given  a file name, build a unique path in the system's temp folder.
    root, extension = os.path.splitext(file_name)
    root = os.path.join(tempfile.gettempdir(), root)

    suffix = 0
    temp_unique = "".join([root, extension])

    while os.path.exists(temp_unique):
        temp_unique = "".join([root, "_", str(suffix), extension])
        suffix += 1

    return temp_unique


def _mp4box_escape_path(file_path):
    # Mp4box does not allow to use file names containing characters '%' and '#', so we replace them with 'rep'.
    reserved_chars = r"[%#]"
    directory, file_name = os.path.split(file_path)
    if re.search(reserved_chars, file_name):
        return _make_temp_unique_path(re.sub(reserved_chars, "rep", file_name))
    return None


def _escape_convert(file_path, file_format):
    root, extension = os.path.splitext(file_path)
    escaped_path = _mp4box_escape_path(file_path)
    extension_lowercase = extension.lower()

    if extension_lowercase in _supported_files_no_conversion:
        if not escaped_path:
            return None

        if _windows:
            # Creating a symlink on Windows is (a) not possible with the standard Python library and (b) requires
            # special privileges to be granted to the client. It's much easier to just make a temporary copy of the
            # file.
            shutil.copy(file_path, escaped_path)
        else:
            os.symlink(file_path, escaped_path)
        return escaped_path

    if extension_lowercase in _supported_files_conversion:
        print(("\nconverting " + file_path + " to " + file_format + "..."))
        sys.stdout.flush()

        _, escaped_name = os.path.split(escaped_path if escaped_path else file_path)
        escaped_root, _ = os.path.splitext(escaped_name)
        escaped_path = _make_temp_unique_path(escaped_root + ".m4a")

        converter = os.path.join(_getProgramPath(), "avconv_win", "avconv.exe") if _windows else "afconvert"
        call_args = [converter]

        if _windows:
            call_args.extend(["-i", file_path])
            if file_format == "aac":
                call_args.extend(["-b", "256k"])
            else:
                call_args.extend(["-c:a", file_format])
        else:
            call_args.extend(["-d", file_format])
            if file_format == "aac":
                call_args.extend(["-b", "256000"])
            call_args.extend([file_path])

        call_args.extend([escaped_path])
        subprocess.check_call(call_args)
        return escaped_path
    else:
        print(("invalid input file format \"" + extension_lowercase + "\""))
        print(("valid input file formats are " + ", ".join(_supported_files_conversion)))
        sys.exit()


def _convert_append(converted_files, file_path, file_format):
    converted_path = _escape_convert(file_path, file_format)
    if converted_path:
        converted_files.append(converted_path)
    return converted_path if converted_path else file_path


def _save_tags(file_path, tags):
    file_tags = mutagen.mp4.Open(file_path)
    if ("track" in tags) and (len(tags["track"]) > 0):
        file_tags["\xa9nam"] = tags["track"]
    if ("artist" in tags) and (len(tags["artist"]) > 0):
        file_tags["\xa9ART"] = tags["artist"]
    if ("release" in tags) and (len(tags["release"]) > 0):
        file_tags["\xa9alb"] = tags["release"]
    if ("remixer" in tags) and (len(tags["remixer"]) > 0):
        file_tags["----:com.apple.iTunes:REMIXER"] = mutagen.mp4.MP4FreeForm(tags["remixer"])
        file_tags["----:com.apple.iTunes:mixartist"] = mutagen.mp4.MP4FreeForm(tags["remixer"])
    if ("mix" in tags) and (len(tags["mix"]) > 0):
        file_tags["----:com.apple.iTunes:MIXER"] = mutagen.mp4.MP4FreeForm(tags["mix"])
    if ("producer" in tags) and (len(tags["producer"]) > 0):
        file_tags["----:com.apple.iTunes:PRODUCER"] = tags["producer"]
    if ("label" in tags) and (len(tags["label"]) > 0):
        file_tags["----:com.apple.iTunes:LABEL"] = mutagen.mp4.MP4FreeForm(tags["label"])
        file_tags["----:com.apple.iTunes:publisher"] = mutagen.mp4.MP4FreeForm(tags["label"])
    if ("genre" in tags) and (len(tags["genre"]) > 0):
        file_tags["\xa9gen"] = tags["genre"]
    if ("track_no" in tags) and (len(tags["track_no"]) > 0):
        if ("track_count" in tags) and (len(tags["track_count"]) > 0):
            file_tags["trkn"] = [(int(tags["track_no"]), int(tags["track_count"]))]
    if ("catalog_no" in tags) and (len(tags["catalog_no"]) > 0):
        file_tags["----:com.apple.iTunes:CATALOGNUMBER"] = mutagen.mp4.MP4FreeForm(tags["catalog_no"])
    if ("year" in tags) and (len(tags["year"]) > 0):
        file_tags["\xa9day"] = tags["year"]
    if ("isrc" in tags) and (len(tags["isrc"]) > 0):
        file_tags["TSRC"] = mutagen.id3.TSRC(encoding=None, text=tags["isrc"])
        file_tags["----:com.apple.iTunes:ISRC"] = mutagen.mp4.MP4FreeForm(tags["isrc"])
    if ("cover" in tags) and (len(tags["cover"]) > 0):
        cover_path = tags["cover"]
        f = urlopen(cover_path)
        file_tags["covr"] = [mutagen.mp4.MP4Cover(f.read(),
          mutagen.mp4.MP4Cover.FORMAT_PNG if cover_path.endswith('png') else
          mutagen.mp4.MP4Cover.FORMAT_JPEG
        )]
        f.close()

    file_tags["----:com.apple.iTunes:AUDIOTYPE"] = mutagen.mp4.MP4FreeForm("STEM".encode("utf-8"))
    file_tags.save(file_path)


def _cleanup_temp_files(temp_file_paths):
    print("\nDeleting files")
    for file_path in temp_file_paths:
        try:
            os.remove(file_path)
        except OSError:
            print(("Could not delete temp file " + file_path))


class StemCreator:

    _defaultMetadata = [
        {"name": "Drums" , "color" : "#FF0000"},
        {"name": "Bass"  , "color" : "#00FF00"},
        {"name": "Synths", "color" : "#FFFF00"},
        {"name": "Other" , "color" : "#0000FF"}
    ]

    def __init__(self, mixdownTrack, stemTracks, fileFormat, metadataFile = None, tags = None):
        self._mixdownTrack = mixdownTrack
        self._stemTracks   = stemTracks
        self._format       = fileFormat if fileFormat else "aac"
        self._tags         = json.load(open(tags)) if tags else {}

        # Mutagen complains gravely if we do not explicitly convert the tag values to a
        # particular encoding. We chose UTF-8, others would work as well.
        for key, value in list(self._tags.items()): self._tags[key] = value.encode('utf-8')

        metaData = [] # ?
        if metadataFile:
            fileObj = codecs.open(metadataFile, encoding="utf-8")
            try:
                metaData = json.load(fileObj)
            except IOError:
                raise
            except Exception:
                raise RuntimeError("Error while reading metadata file")
            finally:
                fileObj.close()

        numStems       = len(stemTracks)
        numMetaEntries = len(metaData["stems"])

        self._metadata = metaData

        # If the input JSON file contains less metadata entries than there are stem tracks, we use the default
        # entries. If even those are not enough, we pad the remaining entries with the following default value:
        # {"name" : "Stem_${TRACK#}", "color" : "#000000"}

        if numStems > numMetaEntries:
            print(("missing stem metadata for stems " + str(numMetaEntries) + " - " + str(numStems)))
            numDefaultEntries = len(self._defaultMetadata)
            self._metadata.extend(self._defaultMetadata["stems"][numMetaEntries:min(numStems, numDefaultEntries)])
            self._metadata["stems"].extend([{"name" :"".join(["Stem_", str(i + numDefaultEntries)]), "color" : "#000000"} \
                for i in range(numStems - numDefaultEntries)])

    def save(self, outputFilePath = None):
        temp_files = []

        try:
            if not outputFilePath:
                root, ext = os.path.splitext(self._mixdownTrack)
                root += ".stem"
            else:
                root, ext = os.path.splitext(outputFilePath)

            outputFilePath = "".join([root, stemOutExtension])
            _removeFile(outputFilePath)

            folderName = "GPAC_win"   if _windows else "GPAC_mac"
            executable = "mp4box.exe" if _windows else "mp4box"
            mp4box     = os.path.join(_getProgramPath(), folderName, executable)

            print("\n[Done 0/6]\n")
            sys.stdout.flush()

            callArgs = [mp4box]
            callArgs.extend(["-add", _convert_append(temp_files, self._mixdownTrack, self._format) + "#ID=Z",
                             outputFilePath])

            print("\n[Done 1/6]\n")
            sys.stdout.flush()

            conversionCounter = 1
            for stemTrack in self._stemTracks:
                callArgs.extend(["-add", _convert_append(temp_files, stemTrack, self._format) + "#ID=Z:disable"])
                conversionCounter += 1
                print(("\n[Done " + str(conversionCounter) + "/6]\n"))
                sys.stdout.flush()
            callArgs.extend(["-udta", "0:type=stem:src=base64," + base64.b64encode(json.dumps(self._metadata).encode()).decode()]) # added byte conversion
            subprocess.check_call(callArgs)
            sys.stdout.flush()

            _save_tags(outputFilePath, self._tags)

            print("\n[Done 6/6]\n")
            sys.stdout.flush()

        except:
            raise
        finally:
            _cleanup_temp_files(temp_files)

        print(("\nCreating " + outputFilePath + " was successful!"))


class StemMetadataViewer:

    def __init__(self, stemFile):
        self._metadata = {}

        if stemFile:
            folderName = "GPAC_win"   if _windows else "GPAC_mac"
            executable = "mp4box.exe" if _windows else "mp4box"
            mp4box     = os.path.join(_getProgramPath(), folderName, executable)

            callArgs = [mp4box]
            callArgs.extend(["-dump-udta", "0:stem", stemFile])
            subprocess.check_call(callArgs)

            root, ext = os.path.splitext(stemFile)
            udtaFile = root + "_stem.udta"
            fileObj = codecs.open(udtaFile, encoding="utf-8")
            fileObj.seek(8)
            self._metadata = json.load(fileObj)
            os.remove(udtaFile)

    def dump(self, metadataFile = None, reportFile = None):
        if metadataFile:
            fileObj = codecs.open(metadataFile, mode="w", encoding="utf-8")
            try:
                fileObj.write(json.dumps(self._metadata))
            except Exception:
                raise
            finally:
                fileObj.close()

        if reportFile:
            fileObj = codecs.open(reportFile, mode="w", encoding="utf-8")
            try:
                for i, value in enumerate(self._metadata["stems"]):
                    line = "Track {:>3}      name: {:>15}     color: {:>8}\n".format(i + 1, value["name"], value["color"])
                    fileObj.write(line)
            except Exception:
                raise
            finally:
                fileObj.close()
